"""Helper functions."""
from __future__ import annotations

import importlib
import inspect

from docstring_parser import parse

from ..generator import MarkdownBuilder


class ClassDocBlock:
    """Class documentation helper"""

    def __init__(self, doc: MarkdownBuilder, module: str, name: str):
        self.doc = doc
        self.class_ = self._get_class(module, name)
        self.render_class(name, self.class_)

    @staticmethod
    def _get_class(module, name):
        module = importlib.import_module(module)
        return getattr(module, name)

    def _render_method(self, obj, attr):
        doc_string = inspect.getdoc(getattr(obj, attr))
        if doc_string:
            parsed = parse(doc_string)
        doc = self.doc
        with doc.heading(f"Method: `{attr}`"):
            spec = str(inspect.signature(getattr(obj, attr)))
            doc.text.paragraph(f"**Signature**: `{spec}`")
            if parsed:
                if parsed.short_description:
                    doc.text.paragraph(parsed.short_description)
                if parsed.long_description:
                    doc.text.paragraph(parsed.long_description)
                with doc.heading("Arguments"):
                    with doc.list.unordered_list() as params:
                        for param in parsed.params:
                            params.append(
                                f"`{param.arg_name}` _({param.type_name})_ - {param.description}"
                            )
                with doc.heading("Returns"):
                    with doc.list.unordered_list(
                        "_(None)_ - Implicit return."
                    ) as returns:
                        for ret in parsed.many_returns:
                            returns.append(f"_({ret.type_name})_ - {ret.description}")

    def _render_attribute(self, obj, attr):
        with self.doc.list.unordered_list() as attrs:
            attrs.append(f"**{attr}**")
            with self.doc.list.unordered_list() as at:
                at.append(f"_Type_: `{getattr(obj, attr).__class__.__name__}`")
                at.append(f"_Default_: `{getattr(obj, attr)}`")

    def render_class(self, name, obj):
        doc = self.doc
        attributes = []
        methods = []
        with self.doc.heading(name):
            doc.text.paragraph(inspect.getdoc(obj))
            for attr in [a for a in dir(obj) if not a.startswith("__")]:
                if callable(getattr(obj, attr)):
                    methods.append((obj, attr))
                if not callable(getattr(obj, attr)):
                    attributes.append((obj, attr))
            if attributes:
                with doc.heading("Attributes"):
                    for attribute in attributes:
                        self._render_attribute(*attribute)
            if methods:
                for method in methods:
                    self._render_method(*method)
            doc.horizontal_bar()



"""Markdown Toolkit main classes."""
from __future__ import annotations

from contextlib import contextmanager
from inspect import cleandoc
from typing import Optional, Union
import itertools
from markdown_toolkit.utils import (
    header,
    list_item,
    sanitise_attribute,
    remove_duplicates,
)


class MarkdownDocument:
    """Markdown document builder class."""

    class _MarkdownList:
        def __init__(
            self,
            item: str,
            document: MarkdownDocument,
            ordered: bool = False,
            prefix: Optional[str] = None,
        ):
            self.doc = document
            self.doc.text(list_item(ordered=ordered, item=item, prefix=prefix))

        def __enter__(self):
            self.doc._indent_level += 1
            self.doc._list_level += 1

        def __exit__(self, exc_type, exc_value, exc_traceback):
            self.doc._indent_level -= 1
            self.doc._list_level -= 1

    class _MarkdownTable:
        """Table renderer."""

        def __init__(
            self,
            document: MarkdownDocument,
            titles: list,
            sort_by: Optional[str] = None,
        ):
            self.doc = document
            self.titles = titles
            self.normalized_titles = list(map(sanitise_attribute, titles))
            self.column_count = len(self.normalized_titles)
            self.rows = []
            self.sort_by = titles.index(sort_by) if sort_by else None

        def bulk_add_rows(self, rows: list[dict]):
            for row in rows:
                row_buffer = []
                for title in self.titles:
                    row_buffer.append(str(row[title]))
                self.rows.append(row_buffer)

        def add_row(self, **columns):
            """Add row to table helper."""
            if columns is None:
                raise ValueError("No data submitted.")
            for column in columns:
                if column not in self.normalized_titles:
                    raise ValueError("Column not found in headers.")
            row_buffer = []
            for title in self.normalized_titles:
                row_buffer.append(str(columns.get(title, "")))
            self.rows.append(row_buffer)

        def _render(self):
            buffer = []
            buffer.append("| " + " | ".join(self.titles) + " |")
            buffer.append(
                "| " + " | ".join(["---" for _ in range(self.column_count)]) + " |"
            )
            if self.sort_by:
                self.rows.sort(key=lambda x: x[self.sort_by])
            for row in self.rows:
                buffer.append("| " + " | ".join(row) + " |")

            return "\n".join(buffer)

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_value, exc_traceback):
            self.doc.paragraph(self._render())

    class _MarkdownHeading:
        """Heading context manager."""

        def __init__(
            self,
            document: MarkdownDocument,
            heading: str,
            silent=False,
            level: Optional[int] = None,
        ):
            self.doc = document
            self.heading = heading
            self.silent = silent
            self.level = level

        def __enter__(self):
            if not self.silent:
                self.doc._buffer.append(self.__str__())
                self.doc.linebreak()
            if not self.level:
                self.doc._heading_level += 1
            return self

        def __exit__(self, exc_type, exc_value, exc_traceback):
            if not self.level:
                self.doc._heading_level -= 1

        def __str__(self) -> str:
            """String representation.

            Returns:
                str: Representation of class as string.
            """
            level = self.level or self.doc._heading_level
            return header(self.heading, level)

    def __init__(self, newline_character: str = "\n"):
        self._buffer: list[str] = []
        self._indent_level: int = -1
        self._list_level: int = -1
        self._heading_level = 1
        self._newline_character: str = newline_character

    @property
    def _in_list(self) -> bool:
        """Helper method to tell if inside a list context.

        Returns:
            bool: In list or not.
        """
        if self._indent_level > -1:
            return True
        return False

    @property
    def _indent(self) -> str:
        """Helper method to indent text in document context.

        Returns:
            str: Indented text.
        """
        indent_muliplier = self._indent_level * 4
        if self._in_list:
            indent_muliplier += 4
        return " " * indent_muliplier

    def heading(
        self,
        heading: Optional[str] = None,
        silent: bool = False,
        level: Optional[int] = None,
    ):
        """_summary_

        Args:
            heading (Optional[str], optional): Text to head with. Defaults to None.
            silent (bool, optional): Increment heading level silently. Defaults to False.
            level (Optional[int], optional): Override heading level. Defaults to None.
        """
        return self._MarkdownHeading(self, heading=heading, silent=silent, level=level)

    def table(
        self,
        raw_table: Optional[list[dict]] = None,
        *,
        titles: Optional[list] = None,
        sort_by: Optional[str] = None,
    ) -> _MarkdownTable:
        """Table rendering helper.

        Args:
            titles (list): List of titles for the columns.

        Returns:
            _MarkdownTable: Object with helper methods.
        """
        if not raw_table:
            return self._MarkdownTable(self, titles=titles, sort_by=sort_by)
        all_titles = remove_duplicates(
            itertools.chain(*[dictionary.keys() for dictionary in raw_table])
        )
        with self._MarkdownTable(self, titles=all_titles, sort_by=sort_by) as table:
            table.bulk_add_rows(raw_table)

    def list(self, item: str, ordered: bool = False, prefix: Optional[str] = None):
        """Returns list context manager, can be used directly."""
        return self._MarkdownList(
            item=item, ordered=ordered, document=self, prefix=prefix
        )

    @contextmanager
    def codeblock(self, language: str = ""):
        """Codeblock context manager."""
        self.text("```" + language)
        yield
        self.text("```")

    @contextmanager
    def indentblock(self):
        """Returns indented context manager."""
        self._indent_level += 1
        yield
        self._indent_level -= 1

    def add(self, text):
        """Unmodified text injection into document."""
        self._buffer.append(text)

    def text(self, text: str = ""):
        """Add text to document, taking into account indent level."""
        self._buffer.append(f"{self._indent}{text}")

    def paragraph(self, text: str, linebreak: Union[int, bool] = True):
        """Add triplequote python multiline strings to document."""
        buffer = cleandoc(text).split("\n")
        for line in buffer:
            self.text(line)
        if linebreak:
            for _ in range(int(linebreak)):
                self.linebreak()

    def linebreak(self):
        """Adds a linebreak to the document."""
        self.add("")

    def horizontal_line(self):
        """Add horizontal line to document."""
        self.text()
        self.text("----")
        self.text()

    @contextmanager
    def injector(self, anchor: str):
        """Add inline comment to use as an injector anchor."""
        self.text(f"<!--- markdown-toolkit:start:{anchor} --->")
        yield
        self.text(f"<!--- markdown-toolkit:end:{anchor} --->")

    def render(self, trailing_whitespace=False) -> str:
        """Renders document to string.

        Returns:
            str: Rendered document.
        """
        document = "\n".join(self._buffer)
        if trailing_whitespace:
            return document + "\n"
        return document



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['markdown_toolkit', 'markdown_toolkit.helpers']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'markdown-toolkit',
    'version': '0.2.0',
    'description': 'Utility package for programmatically creating markdown documents',
    'long_description': '![https://raw.githubusercontent.com/dcurtis/markdown-mark/99572b4a4f71b4ea2b1186a30f440ff2fcf66d27/svg/markdown-mark.svg](https://raw.githubusercontent.com/dcurtis/markdown-mark/99572b4a4f71b4ea2b1186a30f440ff2fcf66d27/svg/markdown-mark.svg)\n# Markdown Toolkit\n\n> **INFO:** This readme is dynamically generated via [`generate_readme.py`](generate_readme.py). **_No changes to this file will be persistent between Github Actions runs._**\n\nA python library for creating and manipulating markdown.\n\nThis library heavily utilises context managers to encapsulate \nlogical blocks in the markdown. Primarily this is used to keep \ntrack of the heading levels, so nested `heading` context\nmanagers will keep track of the header level.\n\n## Examples\n\n> **INFO:** Examples here are automatically taken from the `examples` directory and shown below.\n\n### `examples/example_tables.py`\n\n```python\nfrom markdown_toolkit import MarkdownDocument\n\ndoc = MarkdownDocument()\n# You can use a table as a context manager and lazily add rows\nwith doc.table(titles=["Apple Type", "Grown Count"], sort_by="Grown Count") as table:\n    table.add_row(apple_type="Granny Smith", grown_count=3)\n    table.add_row(apple_type="Golden Delicious", grown_count=1)\n\nwith doc.table(titles=["Ticker", "Stock Percentage"]) as table:\n    table.add_row(ticker="AAPL", stock_percentage="7.14%")\n    table.add_row(ticker="MSFT", stock_percentage="6.1%")\n    table.add_row(ticker="AMZN", stock_percentage="3.8%")\n    table.add_row(ticker="TSLA", stock_percentage="2.5%")\n    table.add_row(ticker="GOOGL", stock_percentage="2.2%")\n\n# Or render a list of dictionaries in a table in one go\nraw_table_data = [\n    {"Major": "Biology", "GPA": "2.4", "Name": "Edward"},\n    {"Major": "Physics", "GPA": "2.9", "Name": "Emily"},\n    {"Major": "Mathematics", "GPA": "3.5", "Name": "Sarah"},\n]\ndoc.table(raw_table_data)\n\nwith open("examples/example_tables.md", "w", encoding="UTF-8") as file:\n    file.write(doc.render())\n\n```\n**_Markdown Output:_**\n`examples/example_tables.md`\n```markdown\n| Apple Type | Grown Count |\n| --- | --- |\n| Golden Delicious | 1 |\n| Granny Smith | 3 |\n\n| Ticker | Stock Percentage |\n| --- | --- |\n| AAPL | 7.14% |\n| MSFT | 6.1% |\n| AMZN | 3.8% |\n| TSLA | 2.5% |\n| GOOGL | 2.2% |\n\n| Major | GPA | Name |\n| --- | --- | --- |\n| Biology | 2.4 | Edward |\n| Physics | 2.9 | Emily |\n| Mathematics | 3.5 | Sarah |\n\n```\n**_Markdown Rendered:_**\n| Apple Type | Grown Count |\n| --- | --- |\n| Golden Delicious | 1 |\n| Granny Smith | 3 |\n\n| Ticker | Stock Percentage |\n| --- | --- |\n| AAPL | 7.14% |\n| MSFT | 6.1% |\n| AMZN | 3.8% |\n| TSLA | 2.5% |\n| GOOGL | 2.2% |\n\n| Major | GPA | Name |\n| --- | --- | --- |\n| Biology | 2.4 | Edward |\n| Physics | 2.9 | Emily |\n| Mathematics | 3.5 | Sarah |\n\n\n----\n\n### `examples/example_headings.py`\n\n```python\nfrom markdown_toolkit import MarkdownDocument\n\ndoc = MarkdownDocument()\nwith doc.heading("Markdown Toolkit"):\n    doc.paragraph("Items added inside the context manager respect the heading level.")\n    with doc.heading("Nested Header"):\n        doc.paragraph(\n            """\n        Lorem ipsum dolor sit amet, consectetur adipiscing elit. \n        Vestibulum ornare magna tincidunt, luctus neque eu, iaculis nisl. \n        Duis porta pharetra hendrerit. Donec ac bibendum purus. Phasellus \n        sagittis tincidunt metus, a condimentum enim malesuada ut. \n        Integer porta pellentesque tempus. Vivamus erat est, imperdiet \n        id enim in, rutrum luctus nisi. Praesent in nibh eleifend, \n        consequat est quis, ultricies dolor. \n        """\n        )\n\nwith open("examples/example_headings.md", "w", encoding="UTF-8") as file:\n    file.write(doc.render())\n\n```\n**_Markdown Output:_**\n`examples/example_headings.md`\n```markdown\n# Markdown Toolkit\n\nItems added inside the context manager respect the heading level.\n\n## Nested Header\n\nLorem ipsum dolor sit amet, consectetur adipiscing elit. \nVestibulum ornare magna tincidunt, luctus neque eu, iaculis nisl. \nDuis porta pharetra hendrerit. Donec ac bibendum purus. Phasellus \nsagittis tincidunt metus, a condimentum enim malesuada ut. \nInteger porta pellentesque tempus. Vivamus erat est, imperdiet \nid enim in, rutrum luctus nisi. Praesent in nibh eleifend, \nconsequat est quis, ultricies dolor. \n\n```\n**_Markdown Rendered:_**\n# Markdown Toolkit\n\nItems added inside the context manager respect the heading level.\n\n## Nested Header\n\nLorem ipsum dolor sit amet, consectetur adipiscing elit. \nVestibulum ornare magna tincidunt, luctus neque eu, iaculis nisl. \nDuis porta pharetra hendrerit. Donec ac bibendum purus. Phasellus \nsagittis tincidunt metus, a condimentum enim malesuada ut. \nInteger porta pellentesque tempus. Vivamus erat est, imperdiet \nid enim in, rutrum luctus nisi. Praesent in nibh eleifend, \nconsequat est quis, ultricies dolor. \n\n\n----\n\n### `examples/example_lists.py`\n\n```python\nfrom markdown_toolkit import MarkdownDocument, link, bold, italic\n\ndoc = MarkdownDocument()\ndoc.list("Example Item")\nwith doc.list("List Item that has children"):\n    doc.list("Child Item")\n    doc.list("Child Item")\n    doc.list("Child Item")\n\ndoc.list("Ordered lists are possible too", ordered=True)\nwith doc.list("And children can be ordered", ordered=True):\n    doc.list("First Child", ordered=True)\n    doc.list("Second Child", ordered=True)\n    with doc.list("Third Child", ordered=True):\n        doc.list("First Grandchild", ordered=True)\n\nwith doc.list("You can nest Paragraphs too!"):\n    doc.linebreak()\n    doc.paragraph(\n        "This is quite helpful when you\'re using lists for document breaks rather than items."\n    )\n    doc.paragraph(\n        """\n        Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum ornare magna \n        tincidunt, luctus neque eu, iaculis nisl. Duis porta pharetra hendrerit. Donec ac \n        bibendum purus. Phasellus sagittis tincidunt metus, a condimentum enim malesuada ut.\n        Integer porta pellentesque tempus. Vivamus erat est, imperdiet id enim in, rutrum\n        luctus nisi. Praesent in nibh eleifend, consequat est quis, ultricies dolor. \n    """\n    )\n    doc.text(\n        link(\n            uri="https://github.com/danielloader/markdown-toolkit",\n            text="Links supported too",\n        )\n    )\n    doc.list("And you can nest children from this too.")\n    doc.list("Should you want to.")\n\n\nwith open("examples/example_lists.md", "w", encoding="UTF-8") as file:\n    file.write(doc.render())\n\n```\n**_Markdown Output:_**\n`examples/example_lists.md`\n```markdown\n*   Example Item\n*   List Item that has children\n    *   Child Item\n    *   Child Item\n    *   Child Item\n1.  Ordered lists are possible too\n1.  And children can be ordered\n    1.  First Child\n    1.  Second Child\n    1.  Third Child\n        1.  First Grandchild\n*   You can nest Paragraphs too!\n\n    This is quite helpful when you\'re using lists for document breaks rather than items.\n\n    Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum ornare magna \n    tincidunt, luctus neque eu, iaculis nisl. Duis porta pharetra hendrerit. Donec ac \n    bibendum purus. Phasellus sagittis tincidunt metus, a condimentum enim malesuada ut.\n    Integer porta pellentesque tempus. Vivamus erat est, imperdiet id enim in, rutrum\n    luctus nisi. Praesent in nibh eleifend, consequat est quis, ultricies dolor. \n\n    [Links supported too](https://github.com/danielloader/markdown-toolkit)\n    *   And you can nest children from this too.\n    *   Should you want to.\n```\n**_Markdown Rendered:_**\n*   Example Item\n*   List Item that has children\n    *   Child Item\n    *   Child Item\n    *   Child Item\n1.  Ordered lists are possible too\n1.  And children can be ordered\n    1.  First Child\n    1.  Second Child\n    1.  Third Child\n        1.  First Grandchild\n*   You can nest Paragraphs too!\n\n    This is quite helpful when you\'re using lists for document breaks rather than items.\n\n    Lorem ipsum dolor sit amet, consectetur adipiscing elit. Vestibulum ornare magna \n    tincidunt, luctus neque eu, iaculis nisl. Duis porta pharetra hendrerit. Donec ac \n    bibendum purus. Phasellus sagittis tincidunt metus, a condimentum enim malesuada ut.\n    Integer porta pellentesque tempus. Vivamus erat est, imperdiet id enim in, rutrum\n    luctus nisi. Praesent in nibh eleifend, consequat est quis, ultricies dolor. \n\n    [Links supported too](https://github.com/danielloader/markdown-toolkit)\n    *   And you can nest children from this too.\n    *   Should you want to.\n\n----\n',
    'author': 'Daniel Loader',
    'author_email': 'hello@danielloader.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/danielloader/markdown-toolkit/',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Markdown Toolkit document injector."""
from __future__ import annotations

import re
from collections import defaultdict
from typing import TextIO

from markdown_toolkit.utils import sanitise_attribute


class DocumentInjector:
    """Class for injected text into Markdown Documents."""

    class Inject:
        """Inject text between start and end lines."""

        def __init__(self, file_buffer, start, end):
            self.file_buffer: list = file_buffer
            self.start = start
            self.end = end

        def read(self) -> str:
            """Return text between anchors."""
            return "".join(self.file_buffer[self.start : self.end - 1])

        def write(self, text: str):
            """Replace text between anchors."""
            del self.file_buffer[self.start : self.end - 1]
            self.file_buffer.insert(self.start, text)

    def __init__(
        self,
        file_obj: TextIO,
    ):

        self.file_buffer: list[str] = file_obj.read().splitlines()
        self._find_tags()
        self.tags = {}

    def render(self, trailing_whitespace=False) -> str:
        """Renders document to string.

        Returns:
            str: Rendered document.
        """
        document = "\n".join(self.file_buffer)
        if trailing_whitespace:
            return document + "\n"
        return document

    def __getattr__(self, attr):
        raise ValueError(attr)

    def _find_tags(self):
        tags = defaultdict(dict)

        for idx, line in enumerate(self.file_buffer):
            if line.startswith("<!--"):
                start = re.match(r"^<!--.*markdown-toolkit:start:(.*).*-->$", line)
                end = re.match(r"^<!--.*markdown-toolkit:end:(.*).*-->$", line)
                if start:
                    anchor = start.groups()[0].strip()
                    if "start" in tags[anchor]:
                        raise ValueError()
                    tags[anchor]["start"] = idx + 1
                if end:
                    anchor = end.groups()[0].strip()
                    if "end" in tags[anchor]:
                        raise ValueError()
                    tags[anchor]["end"] = idx + 1

        ranges = []
        for lines in tags.values():
            ranges.append(set(range(lines["start"], lines["end"])))
        if len(ranges) > 1:
            overlaps = set.intersection(*ranges)
            if overlaps:
                raise IndexError("Overlapping anchors", overlaps)

        self.tags = tags
        for anchor, lines in tags.items():
            setattr(
                self,
                sanitise_attribute(anchor),
                self.Inject(self.file_buffer, lines["start"], lines["end"]),
            )



"""Utilities for inline manipulating strings."""

from inspect import cleandoc
from pathlib import Path
from re import sub
from typing import Optional, Union
from urllib.parse import quote as urlquote

from markdown_toolkit import constants

__all__ = [
    "badge",
    "bold",
    "code",
    "from_file",
    "header",
    "image",
    "italic",
    "link",
    "quote",
    "strikethrough",
]


def sanitise_attribute(string) -> str:
    """Converts any string into a safe python attribute string."""
    return "_".join(
        sub(
            "([A-Z][a-z]+)",
            r" \1",
            sub("([A-Z]+)", r" \1", string.replace("-", " ").replace(".", " ")),
        ).split()
    ).lower()


def from_file(path: Union[Path, str], start: int = 1, end: int = None) -> str:
    """File reader helper.

    Args:
        path (Union[Path,str]): File path to open.
        start (int, optional): Start Line. Defaults to None.
        end (int, optional): End Line. Defaults to None.

    Returns:
        str: Text block.
    """
    with open(Path(path), "r", encoding="UTF-8") as file:
        return "".join(file.readlines()[start - 1 : end])


def badge(label: str, color: str, message: Optional[str] = None, alt: str = "") -> str:
    """Shields.io badge helper.

    Args:
        label (str): Badge label.
        color (str): Badge color.
        message (Optional[str], optional): Badge message. Defaults to None.
        alt (str, optional): Alt tag for the badge. Defaults to "".

    Returns:
        str: _description_
    """
    badge_url = (
        f"https://img.shields.io/static/v1?label="
        f"{urlquote(str(label))}&color={urlquote(str(color))}"
    )
    if message:
        badge_url += f"&message={urlquote(str(message))}"
    return link(uri="https://shields.io/", text=image(uri=badge_url, text=alt))


def list_item(item: str, ordered=False, prefix=None):
    """Returns a list item."""
    if not prefix:
        prefix = constants.ORDERED_LIST if ordered else constants.UNORDERED_LIST
    return f"{prefix.ljust(4)}{cleandoc(item)}"


def quote(text: str, qoute_all_lines=False) -> str:
    """Quotes text."""
    buffer = []
    multiline_text = iter(cleandoc(text).splitlines(keepends=True))
    first = next(multiline_text)
    buffer.append(f"> {first}")

    for line in multiline_text:
        buffer.append(f'{"> " if qoute_all_lines else ""}{line}')
    return "".join(buffer)


def bold(text: str) -> str:
    """Bold wrapper."""
    return f"**{text}**"


def italic(text: str) -> str:
    """Bold wrapper."""
    return f"_{text}_"


def code(text: str) -> str:
    """Code wrapper."""
    return f"`{text}`"


def strikethrough(text: str) -> str:
    """Strikethrough wrapper."""
    return f"~~{text}~~"


def header(heading: str, level: int) -> str:
    """Heading wrapper."""
    return f"{'#'*level} {heading}"


def image(uri: str, *, text: Optional[str] = None, title: Optional[str] = None) -> str:
    """Add an image to the document."""
    return f"!{link(uri, text=text, title=title)}"


def link(uri: str, *, text: Optional[str] = None, title: Optional[str] = None) -> str:
    """Add an link to the document."""
    rendered_link = f"[{text or uri}]({uri}"
    if title:
        rendered_link += f' "{title}"'
    return f"{rendered_link})"


def remove_duplicates(seq):
    seen = set()
    seen_add = seen.add
    return [x for x in seq if not (x in seen or seen_add(x))]
